// LinkerAT - A Utility for cross referencing GIS network databases
//
// Path following cross indexer for networks
//
// developed by: E.F Granzow
//               30 January 2021
//
// This program utilizes a path following algorithm to cross index two geographic networks with dissimilar extents
// and segment structures. Originally developed to link the ARNOLD base network to TMC segments used in HPMRDS, 
// typically applied on a statewide basis. The model uses visit/revisit logic to ensure all corresponding segments
// are matched. To support dissimilar segment structures a "blob" variable is defined to list partial segments and 
// associated milepost distances as a delimited string.
//
// updates: launch and separate powershell driven report windows added (2/10/2022)
//          settings load and save fully implemented (2/27/2022)
//          ms file dialogs added for settings (3/15/2022)
//          draft support for "Arkansas" added and tested (3/16/2022)
//          importfix added to convert prop names to upper case and multilinestring to linestring (3/18/2022)
//          support added to catch input links w/o defined geography (3/23/2022)
//          check and alert added to catch incompatible browser launch (3/25/2022)
//          M/S common file dialog use eliminated - settings now generate settings filename directly (3/1/2023)
//          nircmd command sequence added to eliminate browser refresh problems when monitor windows start (3/20/2023) 
//
// define windows script
//
//   if (!WScript) alert("WScript could not be created - Contact Support");  
   var WshShell = new ActiveXObject("WScript.Shell");
   if (typeof(batchflag)=="undefined") var batchflag=false;
   if (batchflag && dbtrace) WScript.Echo("Entering Main "+WScript.Version); 
//
// set up globals
//
   var report_file;
   var summary_file;
   var settingspath="C:\\LinkerAT\\Settings\\";
   var settingsname;                               // global to hold file name assembled from project description
   var dsettingsf="default.ini";                   // used for initial startup
   var tsettingsf="template.ini";                  // used to get list of settings parameters for save operation
   var setfpath=dsettingsf;
   var chain_file;
   var xidtable;
   var tmctable;
   var webformat;
   var browserName;
//
// define JSON and functions if necessary
//
// callable windows
//
   if (!batchflag)
   {
//      var filewin = new ActiveXObject("MSComDlg.CommonDialog");
//      var dlgdump=JSON.stringify(filewin,Object.getOwnPropertyNames(filewin));
//      if (filewin==undefined) var filewin=null;
//      filewin.MaxFileSize = 260;
      webenable=true
   }
   else webenable=false;
//   filewin.ShowOpen();
//   filewin.ShowSave();
//
// Implementation Notes
//
// !Corridor_ID should be reintialized for first pass 
//
   var clerrors=new Array();    //accumulator for error dump routine
//
// database handles and descriptors
//
   var alinks=new Array();      //name of target link layer
   var tlinks=new Array();      //name of source link layer
//
// spatial indexing (used in tile referencing)
//
   var lindex=new Array();
   var lcxref=new Array();      
//
// "visit" arrays and management
//
//   var avisit=new Array();      //target visit cache
//   var tvisit=new Array();      //source visit cache 
   var tvisits;                 //cumulative source visits 
   var tvisittot;               //number of possible source visits
   var avisits;                 //cumulative target visits
   var avisittot;               //number of possible target visits
   var visit;                   //current index in main loop
//
//internal cross reference caches
//
   var nodeprs=new Array();                 //cache for cross referenced directional ab nodes for source links
   var tmctable=new Array();                //cache for source index to ID lookup 
   var xID=new Array(2);                    //cache for cross reference for discovered source ID-target ID route pairs
   xID[0]=["sourcename"];
   xID[1]=["targetname"];
//
// start, end and error reporting
//
   var slinka=new Array(3);
   var elinka=new Array(3);
   var clerror=new Array(7);
   var excepts=new Array(0,0,0,0,0,0,0);   //exception accumulators for real time updating and final stat report
   var exceptrt=true;                     //flag for real time updating
   var errorset;
//
// these variables controls file naming and calls to local functions (they are normally read from settings)
//
   var root="c:\\";                 //server-drive
   var rootf="ARNOLD-RITIS";                                //project folder
   var state="Colorado";                                    //state whose data is being processed
   var subarea="PuebloCo";                                  //state or defined subregion (e.g. counties)
   if (subarea=="None") subarea=state;                                  //state or defined subregion (e.g. counties)
   var year="2017";                                         //year of source and target data files
   var sourcedf="RITIS";                                    //name of source data subfolder
   var targetdf="ARNOLD";                                   //name of target data subfolder
//
// LinkerAT default file paths
//
   var sourcef=root+"\\"+rootf+"\\"+state+"\\"+subarea+"\\"+year+"\\"+sourcedf+"\\"+subarea+".JSON";
   var targetf=root+"\\"+rootf+"\\"+state+"\\"+subarea+"\\"+year+"\\"+targetdf+"\\"+subarea+".JSON";
   var sourcebf=root+"\\"+rootf+"\\"+state+"\\"+subarea+"\\"+year+"\\"+sourcedf+"\\"+subarea+"_base.JSON";
   var targetbf=root+"\\"+rootf+"\\"+state+"\\"+subarea+"\\"+year+"\\"+targetdf+"\\"+subarea+"_base.JSON";
   var sourceof=root+"\\"+rootf+"\\"+state+"\\"+subarea+"\\"+year+"\\"+sourcedf+"\\"+subarea+"_out.JSON";
   var targetof=root+"\\"+rootf+"\\"+state+"\\"+subarea+"\\"+year+"\\"+targetdf+"\\"+subarea+"_out.JSON";
   var xidf=root+"\\"+rootf+"\\"+state+"\\"+subarea+"\\"+year+"\\Resources/XID.CSV";
   var chainf=root+"\\"+rootf+"\\"+state+"\\"+subarea+"\\"+year+"\\Outputs\\"+subarea+"_"+targetdf+"_Chains.txt";
   var reportf=root+"\\"+rootf+"\\"+state+"\\"+subarea+"\\"+year+"\\Outputs\\"+subarea+"_"+sourcedf+"-"+targetdf+"_Report.txt";
   var summaryf=root+"\\"+rootf+"\\"+state+"\\"+subarea+"\\"+year+"\\Outputs\\"+subarea+"_"+sourcedf+"-"+targetdf+"_Summary.txt";
//
// general parameters used in intialization (available on web page)
// 
   var tracelink=0;             //will interrupt operation at start of processing for debugging 
   var restart=1;               //restart sets start of visits to a user set TMC link index (1=all visits and seeds)
   var rununtil=0;              //optionally specifies last source link to complete before exiting
   var test_index=null;         //use test index to invoke "chain builder" tracing and debugging
   var returntrip=false;         //use current set of VISIT values in source data for inititalization
//   snode_ct = 1     //*** new model drives seeding from file (see file info)
   var buffer=.01;              //search radius for match nodes for unlinked segments (50 ft radius)  
   var arctol=10.0;             //angle difference allowed in start link match
   var disttol=.1               //distance tolerance for using target node as source node surrogate (mi)
   var maxsnap=.02;             //defines maximum snap distance before adding new nodes to represent target nodes
   var buffac=5.0;              //relaxed constraint for buffer size with stronger candidate criteria
   var newnodect=0;             //counts number of link split operations required
   var nchgupd;                 //flags network cfg changes (i.e. node added)
   var browsepage=null;         //location for doucment writes if run through browser (null for command line)
   var statpage=null;           //location for document writes if run through browser (null for command line)
   var tpvalidate=false;        //option to attempt validate link vs flow direction for one way target segments
   var baseSave=false;          //option to write input and modified geojson objects to output
   var CIDReload=true;          //option to reload target CORRIDOR IDs using state handler-false uses those in input file
   var linkerrortext;           //local variable for error type message from lastest link access (written to target DB)
//
// create file system object to support file access
//
   var fso=new ActiveXObject("Scripting.FileSystemObject");
//
// setup complete - get current settings and launch program
//
   if (batchflag && dbtrace) WScript.Echo ("Calling Settings");   
   getsettings(null);
//
// check for batch flag
//
   if (batchflag) webenable=false;
//
// state specific data (currently Colorado)
//
   if (!webenable) launch();

function launch()
//
// this is the main driver for LinkerAT. SA version calls directly. Web version calls with "run" button
//
{
   if (batchflag && dbtrace) WScript.Echo("Entering Launch"); 
//
// if found, deletes open report/summary window
//
   var cmd="c:/linkerat/scripts/cmd/closereports.bat > c:/linkerat/scripts/cmd/launch.log";
   WshShell.Run(cmd,2,true); 
//   var reportf="c:\\linkerat\\scripts\\cmd\\test_log.txt";
//   var summaryf="c:\\linkerat\\scripts\\cmd\\test_sum.txt";
//
// minimize browser window to avoid paint problems
//
   var cmd="c:\\linkerat\\scripts\\cmd\\nircmd win min process msedge.exe >> c:/linkerat/scripts/cmd/launch.log";
   rc=WshShell.Run(cmd,2,true)
//
// start report and summary displays
//
   reportWrite(sourcedf+"-"+targetdf+" Networks Match - Warnings/Messages ("+nowstring()+")\n\n\n");
   if (logwin)
   {
//      var cmd="powershell &{set-ExecutionPolicy -executionpolicy unrestricted -scope process;"+
//                                   "c:\\linkerat\\scripts\\cmd\\report.ps1 -name "+reportf+" -height 28 -width 160 -title \"LinkerAT-Log\"}";
      var cmd="c:/linkerat/scripts/cmd/report.bat "+reportf+" 28  160 scroll LinkerAT-Log"; //> c:/linkerat/scripts/cmd/launch.log";
      var rc=WshShell.Run(cmd,1);
   }
//
   statWrite(true)
   if (sumwin)
   {
//      var cmd="powershell &{set-ExecutionPolicy -executionpolicy unrestricted -scope process;"+
//                                "c:\\linkerat\\scripts\\cmd\\report.ps1 -name "+summaryf+" -height 18 -width 80 -title \"LinkerAT-Summary\"}";
      var cmd="c:/linkerat/scripts/cmd/report.bat "+summaryf+" 18 80 noscroll LinkerAT-Summary"; //>> c:/linkerat/scripts/cmd/launch.log";
      WshShell.Run(cmd,1);
   }
//
// this does performs a windiw restore
//
   var cmd="powershell &{set-ExecutionPolicy -executionpolicy unrestricted -scope process;"+
                        "c:\\linkerat\\scripts\\cmd\\proceed.ps1}";
   WshShell.Run(cmd,2,true);
//
// restore minimized browser
//
   var cmd="c:\\linkerat\\scripts\\cmd\\nircmd win normal process msedge.exe >> c:/linkerat/scripts/cmd/launch.log";
   rc=WshShell.Run(cmd,2,true)
//
// activate log/summary windows in front of browser (not used - causing display problems
/*
   var cmd="c:\\linkerat\\scripts\\cmd\\nircmd win activate title \"LinkerAT-Log\""; 
   WshShell.Run(cmd,2);
   var cmd="c:\\linkerat\\scripts\\cmd\\nircmd win activate title \"LinkerAT-Summary\""; 
   WshShell.Run(cmd,2);
*/
//   showMonitors("Log",false);
//   showMonitors("Summary",false);
//   if (!confirm("Batch Monitors Test ... Continue with Processing?")) return;
//   alert("testing TURF...Distance="+turf.distance(alinks[1].geometry.coordinates[0],alinks[1].geometry.coordinates[alinks[1].geometry.coordinates.length-1],"miles").toString());
//   if (batchflag)
//   {
      initialize();
      parent.settings.history.go(0);
      reportWrite("\n(Information) "+state+"-"+subarea+" Initialization Completed ... Starting Visits\n\n");
      visit();
      if (browserName=="Microsoft Internet Explorer/Edge")
      {
         revisit();
         finish();
      }
/*   else
   {
      var worker=new Worker("c:/linkerat/scripts/linkerat_full.js");
      WshShell.Run(cmd);
   }*/
}
function checkVersion()
//
// verifies browser type and version 
//
{
var nAgt = navigator.userAgent.toLowerCase();
var fullVersion  = ''+parseFloat(navigator.appVersion); 
var majorVersion = parseInt(navigator.appVersion,10);
//
// In MSIE, the true version is after "MSIE" in userAgent
//
   if (nAgt.indexOf("msie")!=-1 || nAgt.indexOf("trident")!=-1 || nAgt.indexOf("edg/")) 
   {
      browserName = "Microsoft Internet Explorer/Edge";
      if(nAgt.indexOf("edg/")) return;
      fullVersion = nAgt.substring(nAgt.indexOf("rv:")+3,nAgt.indexOf(")"));
      if (parseFloat(fullVersion)>=9.0) return;
   }
   
// In Chrome, the true version is after "Chrome" 
/* else if ((verOffset=nAgt.indexOf("Chrome"))!=-1) 
   {
      browserName = "Chrome";
      fullVersion = nAgt.substring(verOffset+7);
   }
*/
   alert ("LinkerAT requires Internet Explorer (Version 9.0 or greater)");
   return;
}